"""
Base classes for low-level implementation
-----------------------------------------
"""

from __future__ import annotations

from typing import Any

import numpy as np

ERROR_MSG = "Base class should not be called directly."


class Models:
    """
    Base class for the core implementation of the ``Models`` class.

    :see: :py:class:`pigreads.Models` for the main interface to the models.

    The single- and double-precision core C++ implementations inherit from this
    base class.
    """

    Nv: int
    "Maximum number of state variables in the models."

    def __init__(self) -> None:
        raise NotImplementedError(ERROR_MSG)

    def __len__(self) -> int:
        """
        Get the number of models.

        :return: Number of models.
        """
        raise NotImplementedError(ERROR_MSG)

    def get_number_definitions(self) -> int:
        """
        Get the number of model definitions.

        :return: Number of model definitions
        """
        raise NotImplementedError(ERROR_MSG)

    def get_key(self, imodel: int) -> str:
        """
        Get the key of the model with the given index.

        :param imodel: Index of the model.
        :return: Key of the model with the given index.
        """
        raise NotImplementedError(ERROR_MSG)

    def get_parameter(self, imodel: int, iparam: int) -> float:
        """
        Get the parameter with the given indices.

        :param imodel: Index of the model.
        :param iparam: Index of the parameter.
        :return: Parameter value.
        """
        raise NotImplementedError(ERROR_MSG)

    def set_parameter(self, imodel: int, iparam: int, value: float) -> None:
        """
        Set the parameter with the given indices.

        :param imodel: Index of the model.
        :param iparam: Index of the parameter.
        :param value: New parameter value.
        """
        raise NotImplementedError(ERROR_MSG)

    def get_block_size(self) -> list[int]:
        """
        Get the local work size for running OpenCL kernels.

        :return: Local work size.
        """
        raise NotImplementedError(ERROR_MSG)

    def set_block_size(self, block_size: list[int]) -> None:
        """
        Set the local work size for running OpenCL kernels.

        :param block_size: Local work size.
        """
        raise NotImplementedError(ERROR_MSG)

    def add(self, key: str, code: str, Nv: int, params: np.ndarray[Any, Any]) -> None:  # pylint: disable=invalid-name
        """
        Select and enable a model with given parameters.

        :param key: The key of the model to be added.
        :param code: OpenCL code for the model.
        :param Nv: Number of variables.
        :param params: Parameter values.
        """
        raise NotImplementedError(ERROR_MSG)

    def weights(
        self,
        dz: Any,
        dy: Any,
        dx: Any,
        mask: np.ndarray[Any, Any],
        diffusivity: np.ndarray[Any, Any],
    ) -> np.ndarray[Any, Any]:
        """
        Calculate the weights for the diffusion term in the reaction-diffusion
        equation.

        :param dz: The distance between points in the z-dimension.
        :param dy: The distance between points in the y-dimension.
        :param dx: The distance between points in the x-dimension.
        :param mask: 3D boolean array encoding which points are inside the medium.
        :param diffusivity: The diffusivity matrix.
        :return: Weight matrix for the diffusion term, A 5D array of shape (1, Nz, Ny, Nx, 19).
        """
        raise NotImplementedError(ERROR_MSG)

    def run(
        self,
        inhom: np.ndarray[Any, Any],
        weights: np.ndarray[Any, Any],
        states: np.ndarray[Any, Any],
        stim_signal: np.ndarray[Any, Any],
        stim_shape: np.ndarray[Any, Any],
        Nt: Any,  # pylint: disable=invalid-name
        dt: Any,
    ) -> np.ndarray[Any, Any]:
        """
        Run a Pigreads simulation.

        :param inhom: A 3D array with integer values, encoding which model to use at each point. \
                Its value is zero for points outside the medium and one or more for points inside. \
                Values larger than zero are used to select one of multiple models: \
                1 for ``models[0]``, 2 for ``models[1]``, etc.
        :param weights: The weights for the diffusion term, see :py:func:`weights`.
        :param states: The initial states of the simulation, a 4D array of shape \
                (Nz, Ny, Nx, Nv).
        :param stim_signal: A 3D array with the stimulus signal at each time point \
                for all variables, with shape (Nt, Ns, Nv).
        :param stim_shape: A 4D array specifying the shape of the stimulus, \
                with shape (Ns, Nz, Ny, Nx).
        :param Nt: The number of time steps to run the simulation for.
        :param dt: The time step size.
        :return: The final states of the simulation, a 4D array of shape (Nz, Ny, Nx, Nv).
        """
        raise NotImplementedError(ERROR_MSG)


__all__ = ["Models"]



#!/usr/bin/env python3
"""
Manually clean up code by removing unneeded divisions by zero
-------------------------------------------------------------

Usage: replace_safe_division.py before.yaml after.yaml
"""

from __future__ import annotations

import re
import sys
from pathlib import Path

OK_TO_REPLACE: dict[str, bool] = {}


def ok_to_replace(term: str) -> bool:
    """
    If a term may be replaced, return True.

    :param term: String representation of a mathematical term.
    :return: Boolean.
    """
    if term not in OK_TO_REPLACE:
        OK_TO_REPLACE[term] = input(
            f"{term}\nCan this term be zero? [y/N] "
        ).strip().lower() in ["", "n", "N"]
    return OK_TO_REPLACE[term]


def find_matching_parentheses(text: str) -> int:
    """
    Find the index of the first ``)`` ignoring all matching pairs of ``()`` until
    then.

    :param text: A string.
    :return: Index of the first matching ``)``.
    """
    stack = []
    for i, char in enumerate(text):
        if char == "(":
            stack.append(i)
        elif char == ")":
            if stack:
                stack.pop()
            else:
                return i
    return -1


def process(text: str) -> str:
    """
    Replace occurrences of safe division with un-safe division if the user gives
    their ok to replace it.

    :param text: String to process.
    :return: Modified string.
    """
    match = re.search(r"/\s*\(\(\s*fabs\s*\(", text)  # )))
    if match:
        i0 = match.start()
        i1 = i0 + text[i0:].find("(")  # )
        i2 = match.end()
        i3 = i2 + find_matching_parentheses(text[i2:])
        if i3 < i2:
            message = "Unmatched parentheses!"
            raise ValueError(message)
        i4 = i1 + 1 + find_matching_parentheses(text[i1 + 1 :])
        if i4 < i1:
            message = "Unmatched parentheses!"
            raise ValueError(message)

        term = text[i2:i3]

        code = text[i1 : i4 + 1].replace(" ", "")
        code_ref = f"((fabs({term})<VERY_SMALL_NUMBER)?(({term}<0.0)?-VERY_SMALL_NUMBER:VERY_SMALL_NUMBER):{term})".replace(
            " ", ""
        )
        if code != code_ref:
            message = "Code:\n"
            message += code
            message = "\nExpected:\n"
            message += code_ref
            message += "\nCode does not have the expected format!"
            raise ValueError(message)

        term = process(term)
        rest = process(text[i4:])

        if ok_to_replace(term):
            text = f"{text[: i1 + 1]}{term}{rest}"
        else:
            text = f"{text[:i4]}{rest}"
    return text


if __name__ == "__main__":
    with Path(sys.argv[1]).open() as fr, Path(sys.argv[2]).open("w") as fw:
        fw.write(process(fr.read()))



#!/usr/bin/env python3
"""
Manually clean up code by renaming variables
--------------------------------------------
"""

from __future__ import annotations

import re
from pathlib import Path

import click


def parse_rules(rule_lines: list[str]) -> list[tuple[re.Pattern[str], str]]:
    """
    Parse regex renaming rules from lines in a file.

    :param rule_lines: list of lines of the form '<pattern> -> <replacement>'
    :return: list of compiled regex patterns and their replacements
    """
    rules = []
    for line in rule_lines:
        stripped = line.strip()
        if not stripped or stripped.startswith("#"):
            continue
        try:
            pattern_str, repl = map(str.strip, stripped.split("->", 1))
            pattern = re.compile(pattern_str)
            rules.append((pattern, repl))
        except Exception as e:
            message = f"Invalid rule line: {line!r}"
            raise ValueError(message) from e
    return rules


def extract_identifiers(code: str) -> set[str]:
    """
    Extract variable-like identifiers from C source code.

    :param code: C code string
    :return: set of variable names
    """
    # This is a heuristic and not a full parser.
    return set(re.findall(r"\b[_a-zA-Z][_a-zA-Z0-9]*\b", code))


def apply_rules(
    identifiers: set[str], rules: list[tuple[re.Pattern[str], str]]
) -> tuple[dict[str, str], set[str]]:
    """
    Apply renaming rules to a set of identifiers.

    :param identifiers: Original variable names
    :param rules: list of (regex, replacement)
    :return: tuple of (mapping of renamed vars, unchanged vars)
    """
    mapping: dict[str, str] = {}
    seen_new_names: dict[str, str] = {}
    unchanged: set[str] = set()

    for ident in identifiers:
        new = ident
        for pattern, repl in rules:
            new_candidate = pattern.sub(repl, new)
            if new_candidate != new:
                new = new_candidate
        if new != ident:
            if new in seen_new_names and seen_new_names[new] != ident:
                message = (
                    f"Conflict: {ident} and {seen_new_names[new]} both map to {new}"
                )
                raise ValueError(message)
            mapping[ident] = new
            seen_new_names[new] = ident
        else:
            unchanged.add(ident)
    return mapping, unchanged


def replace_in_code(code: str, mapping: dict[str, str]) -> str:
    """
    Replace variable names in code using word boundaries.

    :param code: Original code
    :param mapping: dict of old name to new name
    :return: Modified code
    """
    sorted_keys = sorted(mapping.keys(), key=len, reverse=True)
    for old in sorted_keys:
        code = re.sub(rf"\b{re.escape(old)}\b", mapping[old], code)
    return code


@click.command()
@click.argument("c_file", type=click.Path(exists=True, dir_okay=False, path_type=Path))
@click.argument(
    "rule_file", type=click.Path(exists=True, dir_okay=False, path_type=Path)
)
@click.option(
    "--output",
    "-o",
    type=click.Path(dir_okay=False, path_type=Path),
    help="Output file path.",
)
def main(c_file: Path, rule_file: Path, output: Path | None) -> None:
    """
    Rename variables in a C_FILE using regex RULE_FILE.
    """
    code = c_file.read_text()
    rules = parse_rules(rule_file.read_text().splitlines())
    identifiers = extract_identifiers(code)
    mapping, unchanged = apply_rules(identifiers, rules)
    new_code = replace_in_code(code, mapping)

    click.echo("\nRenamed variables:")
    for old, new in sorted(mapping.items()):
        click.echo(f"  {old} -> {new}")

    click.echo("\n✅ Unchanged variables:")
    for name in sorted(unchanged):
        click.echo(f"  {name}")

    out_path = output or c_file.with_suffix(".new")
    out_path.write_text(new_code)
    click.echo(f"\nModified code written to: {out_path}")


if __name__ == "__main__":
    main()



# Pigreads: Python-integrated GPU-enabled reaction diffusion solver
# Copyright (c) 2024 Desmond Kabus. All rights reserved.

"""
Pigreads Python module
----------------------

This Python module is the main interface to set up and run Pigreads simulations
solving the reaction-diffusion equations with OpenCL and NumPy:

.. math::

    \\partial_t \\underline{u}
    =
    \\underline{P} \\nabla \\cdot \\mathbf D \\nabla \\underline{u}
    +
    \\underline{r}(\\underline{u})

for :math:`\\underline{u}(t, \\mathbf x)`, :math:`t\\in[0, T]`, and
:math:`\\mathbf x\\in\\Omega\\subset\\mathbb R^3`, with initial conditions and
no-flux boundary conditions.

The following equations define a simpler example with only two variables,
:math:`\\underline{u} = (u, v)`, with no diffusion in :math:`v`, and
homogeneous and isotropic diffusion:

.. math::

    \\begin{aligned}
    \\partial_t u
    &=
    D \\nabla^2 u
    +
    r_u(u, v)
    \\\\
    \\partial_t v
    &=
    r_v(u, v)
    \\end{aligned}

Pigreads performs the most expensive calculations on graphics cards
using OpenCL, see :py:func:`Models.run` and :py:func:`Models.weights`.
Input and output as well as setting up and interacting with the simulation
should be done in Python, with the exception of adding source terms, so-called
stimulus currents. Pigreads uses the simplistic finite-differences method and
forward-Euler time stepping.

A Pigreads simulation is usually defined in the following steps:
First, define the geometry of the medium. In this example, we use a 2D plane
with 200 points in both x and y::

    import pigreads as pig
    import numpy as np

    R = 10.
    z, y, x = np.mgrid[0:1, -R:R:200j, -R:R:200j]
    dz, dy, dx = pig.deltas(z, y, x)
    r = np.linalg.norm((x, y, z), axis=0)

Pigreads is optimised for three-dimensional space. For
lower-dimensional simulations, set the number of points in additional
dimensions to one, as done above for the z-dimension.

Calculations are performed at all points with periodic boundary conditions.
The integer field ``inhom`` defines which points are inside (1) the medium and outside (0)::

    inhom = np.ones_like(x, dtype=int)
    inhom[r >= R] = 0

Values of inhom larger than zero can be used to select one of multiple models,
i.e., reaction terms :math:`\\underline{r}`. For an ``inhom`` value of 1,
``models[0]`` is used; and ``models[1]`` for a value of 2, etc. One or more of the
available models can be selected using an instance of the :py:class:`Models`
class::

    models = pig.Models()
    models.add("marcotte2017dynamical", beta=1.389)

This class also has a function to create an array of the same shape as ``inhom``
in space but for a given number of frames in time. The first frame is filled
with the appropriate resting values for each model. Initial conditions can then
be set in the 0th frame::

    states = models.resting_states(inhom, Nframes=100)
    states[0, x < -8, 0] = 1
    states[0, y < 0, 1] = 2

Note that states has five indices, in this order: time, z, y, x, state variable.
This indexing is consistently used in Pigreads and NumPy.

The calculation of the diffusion term :math:`\\underline{P} \\nabla \\cdot
\\mathbf D \\nabla \\underline{u}` is implemented as a weighted sum of neighbouring points.
The weights can be calculated using the function :py:func:`weights`, which also requires the
diffusivity_matrix :math:`D` as input, which is set using :py:func:`diffusivity_matrix`::

    diffusivity = pig.diffusivity_matrix(Df=0.03)
    weights = pig.weights(dz, dy, dx, inhom, diffusivity)

Finally, the simulation can be started using :py:func:`run`, which does ``Nt``
forward-Euler steps and only returns the final states after those steps::

    Nt = 200
    dt = 0.025
    for it in range(states.shape[0] - 1):
        states[it + 1] = pig.run(models, inhom, weights, states[it], Nt=Nt, dt=dt)

The 5D array states can now be analysed and visualised, for instance with Matplotlib:

    import matplotlib.pyplot as plt
    plt.imshow(states[-1, 0, :, :, 0])
    plt.show()

Full examples with more sophisticated plotting outputting an MP4 movie, a
progress bar, and stability checks can be found in the ``examples`` folder in
the Git repository of this project.
"""

from __future__ import annotations

from collections.abc import Iterator, MutableMapping
from datetime import datetime
from pathlib import Path
from textwrap import dedent, indent
from typing import Any, ClassVar

import numpy as np

import pigreads.models
from pigreads import core as _core
from pigreads import core1 as _core1
from pigreads import core2 as _core2
from pigreads._version import version as __version__
from pigreads.schema.model import ModelDefinition, ModelEntry

# pylint: disable=too-many-lines


def weights(
    dz: float = 1.0,
    dy: float = 1.0,
    dx: float = 1.0,
    inhom: np.ndarray[Any, Any] | None = None,
    diffusivity: np.ndarray[Any, Any] | float = 1.0,
    double_precision: bool = False,
) -> np.ndarray[Any, Any]:
    """
    Calculate the weights for the diffusion term in the reaction-diffusion
    equation.

    The implementation of this function is defined in :py:meth:`Models.weights`.

    :param dz: The distance between points in the z-dimension, see :py:func:`deltas`.
    :param dy: The distance between points in the y-dimension.
    :param dx: The distance between points in the x-dimension.
    :param inhom: A 3D array with integer values, encoding which model to use at each point. \
            Its value is zero for points outside the medium and one or more for points inside. \
            If ``None``, all points are considered inside the medium.
    :param diffusivity: The diffusivity matrix, see :py:func:`diffusivity_matrix`. \
            If a scalar is given, the matrix is isotropic with the same value in all directions.
    :param double_precision: If ``True``, use double precision for calculations.
    :return: Weight matrix for the diffusion term, A 5D array of shape (1, Nz, Ny, Nx, 19).
    """
    models = Models(double_precision=double_precision)
    return models.weights(dz, dy, dx, inhom=inhom, diffusivity=diffusivity)


def run(
    models: Models,
    inhom: np.ndarray[Any, Any],
    weights: np.ndarray[Any, Any],  # pylint: disable=redefined-outer-name
    states: np.ndarray[Any, Any],
    stim_signal: np.ndarray[Any, Any] | None = None,
    stim_shape: np.ndarray[Any, Any] | None = None,
    Nt: int = 1,  # pylint: disable=invalid-name
    dt: float = 0.001,
    double_precision: bool | None = None,
) -> np.ndarray[Any, Any]:
    """
    Run a Pigreads simulation.

    The implementation of this function is defined in :py:meth:`Models.run`.

    :param models: The models to be used in the simulation, see :py:class:`Models`.
    :param inhom: A 3D array with integer values, encoding which model to use at each point. \
            Its value is zero for points outside the medium and one or more for points inside. \
            Values larger than zero are used to select one of multiple models: \
            1 for ``models[0]``, 2 for ``models[1]``, etc.
    :param weights: The weights for the diffusion term, see :py:func:`weights`.
    :param states: The initial states of the simulation, a 4D array of shape \
            (Nz, Ny, Nx, Nv), see :py:func:`Models.resting_states`.
    :param stim_signal: A 3D array with the stimulus signal at each time point \
            for all variables, with shape (Nt, Ns, Nv). If ``None``, no stimulus is applied.
    :param stim_shape: A 4D array specifying the shape of the stimulus, \
            with shape (Ns, Nz, Ny, Nx). If ``None``, no stimulus is applied
    :param Nt: The number of time steps to run the simulation for.
    :param dt: The time step size.
    :param double_precision: If ``True``, use double precision for calculations.
    :return: The final states of the simulation, a 4D array of shape (Nz, Ny, Nx, Nv).
    """
    if double_precision is not None:
        assert models.double_precision == double_precision, (
            "Chosen precision must match with instance of Models."
        )
    return models.run(
        inhom=inhom,
        weights=weights,
        states=states,
        stim_signal=stim_signal,
        stim_shape=stim_shape,
        Nt=Nt,
        dt=dt,
    )


def get_upper_triangle(
    matrix: np.ndarray[Any, Any],
) -> np.ndarray[Any, Any]:
    """
    Convert a 3x3 matrix to a 6D vector, with the diagonal and upper triangle
    of the matrix as elements in the order xx, yy, zz, yz, xz, xy. Additional
    dimensions are supported, but the last two dimensions must each have size
    3.

    :param matrix: A 3x3 matrix.
    :return: A 6D vector.
    """

    matrix = np.array(matrix)
    assert matrix.ndim >= 2, "matrix must have at least two dimensions"
    assert matrix.shape[-1] == 3, "matrix must be a 3x3 matrix (in the last two axes)"
    assert matrix.shape[-2] == 3, "matrix must be a 3x3 matrix (in the last two axes)"
    assert np.allclose(matrix[..., 1, 2], matrix[..., 2, 1]), (
        "The yz and zy components of matrix need to be the same"
    )
    assert np.allclose(matrix[..., 0, 2], matrix[..., 2, 0]), (
        "The xz and zx components of matrix need to be the same"
    )
    assert np.allclose(matrix[..., 0, 1], matrix[..., 1, 0]), (
        "The xy and yx components of matrix need to be the same"
    )
    triag: np.ndarray[Any, Any] = np.stack(
        (
            matrix[..., 0, 0],  # xx
            matrix[..., 1, 1],  # yy
            matrix[..., 2, 2],  # zz
            matrix[..., 1, 2],  # yz
            matrix[..., 0, 2],  # xz
            matrix[..., 0, 1],  # xy
        ),
        axis=-1,
    )
    return triag


def normalise_vector(
    f: np.ndarray[Any, Any] | list[int] | tuple[int, int, int],
    dtype: type = np.single,
) -> np.ndarray[Any, Any]:
    """
    Normalise a 3D vector to unit length.

    :param f: A 3D vector over space with shape (Nz, Ny, Nx, 3).
    :param dtype: Data type of the arrays, i.e., single or double precision floating point numbers.
    :return: A 5D vector with shape (Nz, Ny, Nx, 3, 1).
    """

    f = np.array(f, dtype=dtype)
    assert isinstance(f, np.ndarray)
    assert f.ndim >= 1, "f must be a 3D vector"
    assert f.ndim <= 4, "too many dimensions for f"
    assert f.shape[-1] == 3, "f must be a 3D vector (in the last axis)"
    while f.ndim < 4:
        f.shape = (1, *f.shape)
    norm = np.linalg.norm(f, axis=-1)
    nonzero = norm > 0
    norm.shape = (*norm.shape, 1)
    f[nonzero] /= norm[nonzero]
    f.shape = (*f.shape, 1)
    assert f.ndim == 5, "f must have 5 dimensions: z, y, x, row, col"
    assert f.shape[-1] == 1, "f must be a 3D column vector"
    assert f.shape[-2] == 3, "f must be a 3D column vector"
    return f


def diffusivity_matrix(
    f: np.ndarray[Any, Any] | list[int] | tuple[int, int, int] | None = None,
    n: np.ndarray[Any, Any] | list[int] | tuple[int, int, int] | None = None,
    Df: np.ndarray[Any, Any] | float = 1.0,  # pylint: disable=invalid-name
    Ds: np.ndarray[Any, Any] | float | None = None,  # pylint: disable=invalid-name
    Dn: np.ndarray[Any, Any] | float | None = None,  # pylint: disable=invalid-name
    dtype: type = np.single,
) -> np.ndarray[Any, Any]:
    """
    Define a diffusivity matrix :math:`\\textbf D` for the reaction-diffusion equation.

    If ``f`` and ``n`` are given, the matrix is defined as:

    .. math::

        \\textbf D = \\textbf D_s \\textbf I + (\\textbf D_f - \\textbf D_s)
        \\textbf f \\textbf f^\\mathrm{T} + (\\textbf D_n - \\textbf D_s)
        \\textbf n \\textbf n^\\mathrm{T}

    :param f: The main direction of diffusion, i.e., the fibre direction. \
            3D vector over space with shape (Nz, Ny, Nx, 3). The last index \
            contains three elements: the x, y, and z component of the vector. \
            Optional if :math:`D_f=D_s=D_n`.
    :param n: The direction of weakest diffusion, i.e., the direction normal to \
            the fibre sheets. A 3D vector over space with shape (Nz, Ny, Nx, 3). The \
            last index contains three elements: the x, y, and z component of the \
            vector. Optional if :math:`D_s=D_n`.
    :param Df: The diffusivity in the direction of the fibres, :math:`\\mathbf{f}`.
    :param Ds: The diffusivity in the fibre sheets, but normal to :math:`\\mathbf{f}`. \
            If ``None``, :math:`D_s` is set to :math:`D_f`.
    :param Dn: The diffusivity in the direction normal to the fibre sheets, \
            i.e., along :math:`\\mathbf{n}`. \
            If ``None``, :math:`D_n` is set to :math:`D_s`.
    :param dtype: Data type of the arrays, i.e., single or double precision floating point numbers.
    :return: A 4D array with shape (Nz, Ny, Nx, 6).

    See also :py:func:`get_upper_triangle` for the convention used for the
    last axis of the output array.
    """

    if Ds is None:
        Ds = Df

    if Dn is None:
        Dn = Ds

    if f is None:
        assert np.allclose(Df, Ds), "If Df!=Ds, f must be given"
        f = [0, 0, 0]

    if n is None:
        assert np.allclose(Ds, Dn), "If Ds!=Dn, n must be given"
        n = [0, 0, 0]

    f = normalise_vector(f, dtype=dtype)
    n = normalise_vector(n, dtype=dtype)
    eye = np.eye(3, dtype=dtype)
    eye.shape = (1, 1, 1, 3, 3)

    D = get_upper_triangle(  # pylint: disable=invalid-name
        Ds * eye
        + (Df - Ds) * f * np.swapaxes(f, -1, -2)
        + (Dn - Ds) * n * np.swapaxes(n, -1, -2)
    )
    assert D.ndim == 4
    assert D.shape[-1] == 6
    return D


class ModelParameters(MutableMapping[str, float]):
    """
    A view into the core implementation of the models
    allowing reading and modifying the parameters of a model.

    :param models: Instance of the models class to link to the core.
    :var models: Instance of the models class to link to the core.
    :vartype models: Models
    :param imodel: Index of the model.
    :var imodel: Index of the model.
    :vartype imodel: int
    """

    def __init__(self, models: Models, imodel: int) -> None:
        self._models: Models = models
        self._imodel: int = imodel
        self._keys: list[str] = list(
            models.get_definition(imodel).all_parameters.keys()
        )

    def _key_to_index(self, key: str) -> int:
        """
        Get the index of the parameter with the given key.

        :param key: Key of the parameter.
        :return: Index of the parameter.
        """
        return next(i for i, k in enumerate(self._keys) if k == key)

    def __getitem__(self, key: str) -> float:
        """
        Get the parameter with the given key.

        :param key: Key of the parameter.
        :return: Parameter value.
        """
        return self._models.get_parameter(self._imodel, self._key_to_index(key))

    def __setitem__(self, key: str, value: float) -> None:
        """
        Set the parameter with the given key.

        :param key: Key of the parameter.
        :param value: New value of the parameter.
        """
        self._models.set_parameter(self._imodel, self._key_to_index(key), value)

    def __delitem__(self, key: str) -> None:
        """
        Delete the parameter with the given key.

        Note: This operation is not supported in this class.

        :param key: Key of the parameter.
        """
        message = "Deleting items is not supported in this class."
        raise NotImplementedError(message)

    def __iter__(self) -> Iterator[str]:
        """
        Get an iterator of the keys of the parameters.

        :return: Iterator of the keys of the parameters.
        """
        return iter(self._keys)

    def __len__(self) -> int:
        """
        Get the number of parameters.

        :return: Number of parameters.
        """
        return len(self._keys)

    def to_dict(self) -> dict[str, float]:
        """
        Convert the parameters to a dictionary.

        :return: Dictionary of the parameters.
        """
        return {k: self[k] for k in self._keys}

    def __repr__(self) -> str:
        """
        Get the string representation of the parameters.

        :return: String representation of the parameters.
        """
        return repr(self.to_dict())

    def __str__(self) -> str:
        """
        Get the string representation of the parameters.

        :return: String representation of the parameters.
        """
        return str(self.to_dict())


class Models:
    """
    This class stores the models to be used in a Pigreads simulation. The
    models are defined by OpenCL code. The class variable :py:attr:`available`
    is a dictionary of all available models.

    Models are added to an instance of this class using the :py:meth:`add`
    method, or as a list of tuples in the constructor. The order of the models
    is the order in which they are used in the simulation. ``models[0]`` is used
    at ``inhom`` values of 1, ``models[1]`` at 2, etc.

    :param tuples_key_kwargs: A list of tuples with the model key and keyword \
    arguments encoding parameter names and values to be passed to the \
    :py:meth:`add` method. If a string is given, it is treated the name of the \
    first model to be added with no keyword arguments. If ``None``, no \
    models are added.

    :param double_precision: If ``True``, use double precision for calculations.
    :var double_precision: If ``True``, use double precision for calculations.
    :vartype double_precision: bool
    """

    available: ClassVar[dict[str, ModelDefinition]] = {
        k: ModelDefinition(**v) for k, v in pigreads.models.available.items()
    }
    """
    Dictionary of all available models.
    """

    def __init__(
        self,
        tuples_key_kwargs: list[tuple[str, dict[str, float]]] | str | None = None,
        double_precision: bool = False,
    ):
        self.double_precision: bool = double_precision
        self._core: _core.Models | None = None

        if tuples_key_kwargs is None:
            tuples_key_kwargs = []
        elif isinstance(tuples_key_kwargs, str):
            tuples_key_kwargs = [(tuples_key_kwargs, {})]
        for key, kwargs in tuples_key_kwargs:
            self.add(key, **kwargs)

    @property
    def core(self) -> _core.Models:
        """
        Get the link to the core implementation of this class.
        """
        if self.double_precision:
            if not isinstance(self._core, _core2.Models2):
                self._core = _core2.Models2()
        elif not isinstance(self._core, _core1.Models1):
            self._core = _core1.Models1()
        return self._core

    def __len__(self) -> int:
        """
        The number of models added to the list of models.
        """
        return len(self.core)

    def get_key(self, imodel: int) -> str:
        """
        Get the key of the model with the given index.

        :param imodel: Index of the model.
        """
        return self.core.get_key(imodel)

    def get_definition(self, imodel: int) -> ModelDefinition:
        """
        Get the definition of the model with the given index.

        :param imodel: Index of the model.
        """
        return self.available[self.get_key(imodel)]

    def get_parameter(self, imodel: int, iparam: int) -> float:
        """
        Get the parameter with the given indices.

        :param imodel: Index of the model.
        :param iparam: Index of the parameter.
        :return: Parameter value.
        """
        return self.core.get_parameter(imodel, iparam)

    def set_parameter(self, imodel: int, iparam: int, value: float) -> None:
        """
        Set the parameter with the given indices.

        :param imodel: Index of the model.
        :param iparam: Index of the parameter.
        :param value: New parameter value.
        """
        self.core.set_parameter(imodel, iparam, value)

    def get_entry(self, imodel: int) -> ModelEntry:
        """
        Get the model entry with the given index linked to the core
        implementation to read and change model parameters.

        :param imodel: Index of the model.
        :return: Model entry with a :py:class:`ModelParameters` view.
        """
        model = ModelEntry(key=self.get_key(imodel))
        model.parameters = ModelParameters(models=self, imodel=imodel)
        return model

    def __getitem__(self, imodel: int) -> ModelEntry:
        """
        Get the model entry with the given index linked to the core
        implementation to read and change model parameters.

        :param imodel: Index of the model.
        :return: Model entry with a :py:class:`ModelParameters` view.
        """
        return self.get_entry(imodel)

    def __iter__(self) -> Iterator[ModelEntry]:
        """
        Get an iterator of model entries linked to the core
        implementation to read and change model parameters.

        :return: Iterator of model entries.
        """
        for imodel in range(len(self)):
            yield self.get_entry(imodel)

    @property
    def block_size(self) -> list[int]:
        """
        Local work size for running OpenCL kernels.

        The ``block_size`` parameter corresponds to OpenCL's ``localWorkSize`` and
        defines the size of a small block of the domain that is processed together by
        the OpenCL platform (typically the GPU).

        If chosen too small, the performance will be suboptimal. If chosen too
        large, the OpenCL kernel will not run. The default value is ``(1, 8, 8)``,
        i.e., a block of unit width in z, and ``8x8`` in the y and x
        dimensions. This is a good compromise for most applications.

        Tweaking this value can lead to significant performance improvements.
        """
        return self.core.get_block_size()

    @block_size.setter
    def block_size(self, block_size: list[int]) -> None:
        """
        Set the local work size for running OpenCL kernels.

        :param block_size: Local work size.
        """
        self.core.set_block_size(block_size)

    def add(self, key: str, **parameters: Any) -> None:
        """
        Select and enable a model with given parameters.

        :param key: The key of the model to be added.
        :param parameters: Parameter names and their values to be passed to the model.
        """
        model_id = self.core.get_number_definitions()
        model_def = self.available[key]
        self.core.add(
            key,
            self.code(key, model_id),
            len(model_def.variables),
            np.array(list(model_def(**parameters).values())),
        )

    def code(self, key: str, model_id: int) -> str:
        """
        OpenCL kernel source code defining a model.

        :param key: The key of the model.
        :param model_id: An integer identifying the model.
        """
        model_def: ModelDefinition = self.available[key]
        code: str = ""
        code += dedent(rf"""
            void Model_{key}_step(
                    Real* const params,
                    struct States weights,
                    struct States states_old,
                    struct States states_new,
                    const Real dt
            ) {{
        """).strip()

        code += "\n"

        for ip, param in enumerate(model_def.all_parameters):
            code += f"  const Real {param} = params[{ip}];\n"

        for ivar, varname in enumerate(model_def.variables):
            code += f"  const Real {varname} = _r(_v({ivar}, states_old));\n"
            code += f"  Real* const _new_{varname} = _pr(_v({ivar}, states_new));\n"
            if varname in model_def.diffusivity:
                code += (
                    f"  const Real _diffuse_{varname} = diffusivity_{varname} "
                    + f"* diffuse(weights, _v({ivar}, states_old));\n"
                )

        code += (
            "\n"
            + indent(model_def.get_code(self.double_precision).strip(), "  ")
            + "\n"
        )
        code += dedent(rf"""
            }}

            __kernel void Model_{key}_kernel(
                    Size model_count,
                    __global Size* model_ids,
                    __global Size* model_offsets,
                    __global Real* model_params,
                    __global void* inhom_data,      struct StatesIdx inhom_idx,
                    __global void* weights_data,    struct StatesIdx weights_idx,
                    __global void* states_old_data, struct StatesIdx states_old_idx,
                    __global void* states_new_data, struct StatesIdx states_new_idx,
                    const Real dt
            ) {{

              struct States inhom      = {{inhom_data,      STATES_UNPACK(inhom_idx)}};
              struct States weights    = {{weights_data,    STATES_UNPACK(weights_idx)}};
              struct States states_old = {{states_old_data, STATES_UNPACK(states_old_idx)}};
              struct States states_new = {{states_new_data, STATES_UNPACK(states_new_idx)}};

              const Size iz = get_global_id(0);
              const Size iy = get_global_id(1);
              const Size ix = get_global_id(2);

              if (ix < states_old.Nx && iy < states_old.Ny && iz < states_old.Nz) {{
                const Int inhom_zyx = _i(States_offset(inhom, 0, iz, iy, ix, 0));
                if (inhom_zyx > 0) {{
                  const Size imodel = (inhom_zyx - 1) % model_count;
                  const Size model_id = model_ids[imodel];
                  if (model_id == {model_id}) {{
                    Real* params = model_params + model_offsets[imodel];
                    struct States w = States_offset(weights, 0, iz, iy, ix, 0);
                    struct States u = States_offset(states_old, 0, iz, iy, ix, 0);
                    struct States u_ = States_offset(states_new, 0, iz, iy, ix, 0);
                    Model_{key}_step(params, w, u, u_, dt);
                  }}
                }}
              }}
            }}
        """).strip()

        return code

    @property
    def Nv(self) -> int:  # pylint: disable=invalid-name
        """
        Maximum number of state variables in the models.
        """
        return self.core.Nv

    @property
    def dtype(self) -> type:
        """
        Data type to use for calculations,
        i.e., single or double precision floating point numbers.
        """
        dtype: type = np.double if self.double_precision else np.single
        return dtype

    def resting_states(
        self,
        inhom: np.ndarray[Any, Any],
        Nframes: int = 1,  # pylint: disable=invalid-name
        dtype: type | None = None,
    ) -> np.ndarray[Any, Any]:
        """
        Create an array of states and fill the first frame with the resting
        values of the models depending on the ``inhom`` values.

        :param inhom: A 3D array with integer values, encoding which model to use at each point. \
                Its value is zero for points outside the medium and one or more for points inside. \
                Values larger than zero are used to select one of multiple models: \
                1 for ``models[0]``, 2 for ``models[1]``, etc.
        :param Nframes: The number of frames in time.
        :param dtype: Data type of the arrays, \
                i.e., single or double precision floating point numbers.
        :return: A 5D array of shape (Nframes, Nz, Ny, Nx, Nv).
        """

        if dtype is None:
            dtype = self.dtype

        model_count = len(self)
        assert model_count > 0, "must add at least one model"
        assert inhom.ndim == 3
        inhom = inhom.astype(int)
        mask = inhom > 0
        states: np.ndarray[Any, Any] = np.full(
            (Nframes, *inhom.shape, self.Nv), np.nan, dtype=dtype
        )
        for imodel in range(len(self)):
            model_def = self.get_definition(imodel)
            for iv, resting in enumerate(model_def.variables.values()):
                states[0, mask * ((inhom - 1) % model_count == imodel), iv] = resting
        states[:, ~mask, :] = np.nan
        return states

    def weights(
        self,
        dz: float = 1.0,
        dy: float = 1.0,
        dx: float = 1.0,
        inhom: np.ndarray[Any, Any] | None = None,
        diffusivity: np.ndarray[Any, Any] | float = 1.0,
    ) -> np.ndarray[Any, Any]:
        """
        Calculate the weights for the diffusion term in the reaction-diffusion
        equation.

        :param dz: The distance between points in the z-dimension, see :py:func:`deltas`.
        :param dy: The distance between points in the y-dimension.
        :param dx: The distance between points in the x-dimension.
        :param inhom: A 3D array with integer values, encoding which model to use at each point. \
                Its value is zero for points outside the medium and one or more for points inside. \
                If ``None``, all points are considered inside the medium.
        :param diffusivity: The diffusivity matrix, see :py:func:`diffusivity_matrix`. \
                If a scalar is given, the matrix is isotropic with the same value in all directions.
        :return: Weight matrix for the diffusion term, A 5D array of shape (1, Nz, Ny, Nx, 19).
        """

        assert dz > 0
        assert dy > 0
        assert dx > 0

        if inhom is None:
            inhom = np.ones(shape=(1, 1, 1), dtype=self.dtype)
        assert inhom.ndim == 3

        mask = inhom > 0
        mask.shape = (1, *mask.shape, 1)

        diffusivity = np.array(diffusivity, dtype=self.dtype)
        assert isinstance(diffusivity, np.ndarray)
        if diffusivity.ndim == 0:
            diffusivity = diffusivity_matrix(
                Df=float(diffusivity.item()), dtype=self.dtype
            )
        assert diffusivity.ndim == 4
        assert diffusivity.shape[-1] == 6

        return self.core.weights(dz, dy, dx, mask, diffusivity)

    def run(
        self,
        inhom: np.ndarray[Any, Any],
        weights: np.ndarray[Any, Any],  # pylint: disable=redefined-outer-name
        states: np.ndarray[Any, Any],
        stim_signal: np.ndarray[Any, Any] | None = None,
        stim_shape: np.ndarray[Any, Any] | None = None,
        Nt: int = 1,  # pylint: disable=invalid-name
        dt: float = 0.001,
    ) -> np.ndarray[Any, Any]:
        """
        Run a Pigreads simulation.

        :param inhom: A 3D array with integer values, encoding which model to use at each point. \
                Its value is zero for points outside the medium and one or more for points inside. \
                Values larger than zero are used to select one of multiple models: \
                1 for ``models[0]``, 2 for ``models[1]``, etc.
        :param weights: The weights for the diffusion term, see :py:func:`weights`.
        :param states: The initial states of the simulation, a 4D array of shape \
                (Nz, Ny, Nx, Nv), see :py:func:`Models.resting_states`.
        :param stim_signal: A 3D array with the stimulus signal at each time point \
                for all variables, with shape (Nt, Ns, Nv). If ``None``, no stimulus is applied.
        :param stim_shape: A 4D array specifying the shape of the stimulus, \
                with shape (Ns, Nz, Ny, Nx). If ``None``, no stimulus is applied
        :param Nt: The number of time steps to run the simulation for.
        :param dt: The time step size.
        :return: The final states of the simulation, a 4D array of shape (Nz, Ny, Nx, Nv).
        """
        assert Nt > 0
        assert dt > 0
        assert len(self) > 0, "must add at least one model"
        assert inhom.ndim == 3

        if stim_signal is None or getattr(stim_signal, "size", 0) == 0:
            stim_signal = np.zeros((0, 0, 0, 0, 0), dtype=self.dtype)
        else:  # np.ndarray
            assert stim_signal.ndim in [2, 3]
            stim_signal = np.reshape(
                stim_signal, (stim_signal.shape[0], -1, 1, 1, self.Nv)
            )

        assert isinstance(stim_signal, np.ndarray)
        Ns = stim_signal.shape[1]  # pylint: disable=invalid-name

        if stim_shape is None or getattr(stim_shape, "size", 0) == 0:
            stim_shape = np.zeros((0, 0, 0, 0, 0), dtype=self.dtype)
        else:  # np.ndarray
            assert stim_shape.ndim in [3, 4]
            stim_shape = np.where(inhom > 0, stim_shape, 0)
            stim_shape.shape = (Ns, *stim_shape.shape[-3:], 1)  # type: ignore[union-attr]

        assert isinstance(stim_shape, np.ndarray)
        assert stim_shape.shape[0] == stim_signal.shape[1]

        return self.core.run(
            np.reshape(inhom, (*inhom.shape, 1)),
            weights,
            states,
            stim_signal,
            stim_shape,
            Nt,
            dt,
        )


def to_ithildin(
    framedur: float,
    dt: float,
    dz: float,
    dy: float,
    dx: float,
    models: Models,
    states: np.ndarray[Any, Any],
    inhom: np.ndarray[Any, Any] | None = None,
) -> tuple[dict[str, Any], dict[str, np.ndarray[Any, Any]]]:
    """
    Convert the output of a Pigreads simulation to an Ithildin SimData object.

    While originally designed for a different reaction-diffusion solver,
    the Python module for Ithildin is useful to analyse Pigreads simulations.

    :param framedur: The duration between subsequent frames, usually in milliseconds.
    :param dt: The time step size used in :py:func:`run`, usually in milliseconds.
    :param dz: The distance between points in the z-dimension, see :py:func:`deltas`.
    :param dy: The distance between points in the y-dimension.
    :param dx: The distance between points in the x-dimension.
    :param models: The models used in the simulation, see :py:class:`Models`.
    :param states: The states of the simulation, a 5D array of shape \
            (Nt, Nz, Ny, Nx, Nv), see :py:func:`Models.resting_states` and :py:func:`run`.
    :param inhom: A 3D array with integer values, encoding which model to use at each point. \
            Its value is zero for points outside the medium and one or more for points inside. \
            If ``None``, all points are considered inside the medium.
    :return: Tuple of an Ithildin log file as a string \
            and a dictionary of the variables by variable name.

    Usage::

        import ithildin as ith
        log, variables = pig.to_ithildin(Nt * dt, dt, dz, dy, dx, models, states, inhom)
        sd = ith.SimData(log=ith.Log(data=log))
        sd.vars = variables
    """

    Nt, Nz, Ny, Nx, _ = states.shape  # pylint: disable=invalid-name

    timestamp = datetime.now()

    log = {
        "Ithildin log version": 2,
        "Simulation parameters": {
            "Ithildin library version": f"pigreads {__version__}",
            "Timestep dt": dt,
            "Frame duration": framedur,
            "Number of frames to take": Nt,
            "Serial number": int(timestamp.strftime(r"%Y%m%d%H%M%S")),
            "Name of simulation series": "pigreads",
        },
        "Geometry parameters": {
            "Number of dimensions": 3,
            "Voxel size": [dx, dy, dz],
            "Domain size": [Nx, Ny, Nz],
        },
        "Start date": timestamp.isoformat(),
    }

    for i, model in enumerate(models):
        model_def = models.available[model.key]
        key = "Model parameters"
        if i > 0:
            key += f" {i}"
        log[key] = {
            "Model type": model_def.name,
            "Class": model.key,
            "Citation": "\n".join(model_def.dois),
            "Parameters": model_def(**model.parameters),
            "Initial values": model_def.variables,
            "Variable names": list(model_def.variables.keys()),
            "Number of vars": len(model_def.variables),
        }

    shape = (-1, Nz, Ny, Nx)
    variables: dict[str, np.ndarray[Any, Any]] = {
        v: states[..., iv].reshape(shape)
        for iv, v in enumerate(models.available[models[0].key].variables.keys())
    }
    if inhom is not None:
        variables["inhom"] = inhom.reshape(shape)

    return log, variables


def delta(x: np.ndarray[Any, Any], ax: int = -1) -> float:
    """
    Extract grid spacing from a 3D array.

    :param x: A 3D array.
    :param ax: The axis along which to calculate the distance.
    :return: The distance between the first two points.

    For example, consider this code::

        z, y, x = np.mgrid[0, 0:4:0.2, 0:1:5j]
        dx = pig.delta(z, ax=-1)
        dy = pig.delta(z, ax=-2)
        dz = pig.delta(z, ax=-3)
    """
    assert x.ndim == 3
    diff = np.diff(np.moveaxis(x, ax, -1)[0, 0, :2])
    return 1.0 if diff.shape[0] == 0 else float(diff[0])


def deltas(*x: np.ndarray[Any, Any]) -> list[float]:
    """
    Extract grid spacing from a 3D meshgrid.

    For example, consider this code::

        z, y, x = np.mgrid[0, 0:4:0.2, 0:1:5j]
        dz, dy, dx = pig.deltas(z, y, x)

    :param x: A 3D array.
    :return: A list with the distances between the points.
    """
    return [delta(xi, i) for i, xi in enumerate(x)]


def prepare_array(
    shape: tuple[int, ...],
    path: Path | str | None = None,
    dtype: type = np.single,
) -> np.ndarray[Any, Any]:
    """
    Prepare an array in a given shape.

    Either create a new array or load an existing array from the file with
    the given path as a memory map.

    The shape and dtype of the array are given as arguments. If the path is
    ``None``, a new array is created. If the path is a file, the array is
    loaded from the file. If the array is not of the correct shape or dtype
    or the file does not exist, a new array is created.

    The array is returned as a memory map if a path is given, otherwise as a
    normal numpy array.

    :param shape: Shape of the array.
    :param path: Path to the file to load the array from.
    :param dtype: Data type of the arrays, i.e., single or double precision floating point numbers.
    :return: Resulting array.
    :see: :py:func:`numpy.lib.format.open_memmap`
    """

    if path is None:
        return np.zeros(shape=shape, dtype=dtype)

    path = Path(path)
    if path.is_file():
        arr = np.lib.format.open_memmap(path, "r+")  # type: ignore[no-untyped-call]
        if isinstance(arr, np.ndarray) and arr.shape == shape and arr.dtype == dtype:
            return arr

    arr = np.lib.format.open_memmap(  # type: ignore[no-untyped-call]
        path,
        "w+",
        dtype=dtype,
        shape=shape,
    )
    assert isinstance(arr, np.ndarray)
    arr[:] = np.nan
    return arr


__all__ = [
    "ModelParameters",
    "Models",
    "__version__",
    "delta",
    "deltas",
    "diffusivity_matrix",
    "get_upper_triangle",
    "normalise_vector",
    "prepare_array",
    "run",
    "to_ithildin",
    "weights",
]



"""
Type annotations for double-precision implementation
----------------------------------------------------
"""

from __future__ import annotations

from pigreads.core import Models

# pylint: disable=abstract-method

class Models2(Models):
    """
    Double-precision implementation of the ``Models`` class.

    :see: :py:class:`pigreads.Models` for the main interface to the models.
    """

__all__ = ["Models2"]



"""
Implementation of the CLI
-------------------------
"""

from __future__ import annotations

import json
from typing import Any, TextIO

import click
import numpy as np
import yaml

import pigreads.plot as pp
from pigreads import __version__
from pigreads.progress import PROGRESS_ITERS
from pigreads.schema import Simulation


@click.group()
def main() -> None:
    "Run and visualise Pigreads simulations."


@main.command()
def version() -> None:
    "Show installed version."
    click.echo(f"Pigreads {__version__}")


@main.command()
@click.argument("config", type=click.File("r"))
@click.argument("result", type=click.Path())
@click.option(
    "-p",
    "--progress",
    type=click.Choice(list(PROGRESS_ITERS.keys())),
    default="bar",
    help="Progress bar type.",
)
@click.option(
    "-s", "--start-frame", type=int, default=0, help="Index of frame to start with."
)
@click.option(
    "--live",
    type=str,
    default=None,
    help="Enable live view for a specific variable, given as variable name or index.",
)
@click.option("--live-dpi", type=int, default=None, help="Dots per inch for live view.")
@click.option(
    "--live-path",
    type=str,
    default=None,
    help="File path for live view, if not given, live view is shown in a window.",
)
@click.option(
    "--live-index-z", type=int, default=None, help="Index in z direction for live view."
)
@click.option(
    "--live-kwargs",
    type=yaml.safe_load,
    default="{}",
    help="Additional arguments as YAML or JSON to pass to plot_frame, imshow, and LiveView.",
)
def run(
    config: TextIO,
    result: click.Path,
    progress: str,
    start_frame: int,
    live: str | None,
    live_dpi: int,
    live_path: str | None,
    live_index_z: int | None,
    live_kwargs: dict[str, Any],
) -> None:
    """
    Run a simulation.

    CONFIG is a YAML or JSON file defining the simulation,
    and RESULT is a path to write the results to.
    """

    sim = Simulation(**yaml.safe_load(config))

    liveview = (
        pp.LiveView(
            sim=sim,
            variable=live,
            dpi=live_dpi,
            index_z=live_index_z,
            path=live_path,
            **live_kwargs,
        )
        if live
        else None
    )

    def callback(states: np.ndarray[Any, Any], ifr: int) -> None:
        if liveview:
            liveview.update(states, ifr)

    sim.run(
        path=str(result),
        progress=PROGRESS_ITERS[progress],
        start_frame=start_frame,
        callback=callback,
    )

    if liveview:
        liveview.close()


@main.command()
@click.argument("input", type=click.File("r"), default="-")
@click.argument("output", type=click.File("w"), default="-")
@click.option(
    "-f",
    "--format",
    type=click.Choice(["yaml", "json"]),
    default="yaml",
    help="Output format.",
)
def convert(
    input: TextIO,  # pylint: disable=redefined-builtin
    output: TextIO,
    format: str,  # pylint: disable=redefined-builtin
) -> None:
    """
    Convert a configuration file to canonical format.

    INPUT is a YAML or JSON file defining the simulation,
    and OUTPUT is a path to write the converted file to.
    """
    sim = Simulation(**yaml.safe_load(input))
    data = sim.model_dump()

    if format == "yaml":
        yaml.safe_dump(data, output, allow_unicode=True)
    else:
        assert format == "json"
        json.dump(data, output, indent=2)


@main.command()
@click.argument("config", type=click.File("r"))
@click.argument("result", type=click.Path(exists=True))
@click.argument("movie", type=click.Path())
@click.option("-d", "--dpi", type=int, default=180, help="Dots per inch.")
@click.option("-r", "--fps", type=int, default=15, help="Frames per second.")
@click.option(
    "-p",
    "--progress",
    type=click.Choice(list(PROGRESS_ITERS.keys())),
    default="bar",
    help="Progress bar type.",
)
@click.option("-z", "--index-z", type=int, default=None, help="Index in z direction.")
@click.option(
    "-v", "--variable", type=str, default=None, help="Variable name or index."
)
@click.option(
    "--kwargs",
    type=yaml.safe_load,
    default="{}",
    help="Additional arguments as YAML or JSON to pass to movie, "
    "plot_frame, and imshow, for example "
    '\'{"cmap": "inferno", "vmin": -0.25, "vmax": 1.25}\'.',
)
@click.option(
    "-n",
    "--parallel",
    type=int,
    default=1,
    help="Number of processes (default 1, 0 to use all CPUs).",
)
def movie(
    config: TextIO,
    result: click.Path,
    movie: click.Path,  # pylint: disable=redefined-outer-name
    dpi: int,
    fps: int,
    progress: str,
    variable: str | None,
    index_z: int | None,
    parallel: int,
    kwargs: dict[str, Any],
) -> None:
    """
    Generate a movie from a result file using FFmpeg and Matplotlib.

    CONFIG is a YAML or JSON file defining the simulation,
    RESULT is a NPY file containing the simulation results,
    and MOVIE is a path to write an MP4 video to.
    """

    sim = Simulation(**yaml.safe_load(config))
    models = sim.prepare_models()

    array: np.ndarray[Any, Any] = np.lib.format.open_memmap(  # type: ignore[no-untyped-call]
        result,
        mode="r",
    )
    assert array.shape == (sim.Nfr, sim.Nz, sim.Ny, sim.Nx, models.Nv)
    iz, iv = pp.get_iz_iv(sim, models, variable, index_z)
    array = array[:, iz, :, :, iv]

    zlabel = f" z = {sim.dz * iz:.2f}" if sim.Nz > 1 else ""
    tlables = [
        f"t = {time}" + zlabel for time in (sim.dt * sim.Nt * np.arange(sim.Nfr))
    ]

    pp.movie(
        path=str(movie),
        frames=array,
        tlables=tlables,
        progress=progress,
        parallel=parallel,
        **pp.imshow_defaults(
            sim=sim,
            array=array,
            dpi=dpi,
            fps=fps,
            vlabel=list(models.available[models[0].key].variables)[iv],
            **kwargs,
        ),
    )



"""
Models
------
"""

from __future__ import annotations

from collections.abc import MutableMapping
from typing import Any

from pydantic import BaseModel, ConfigDict, Field, model_validator


class ModelEntry(BaseModel):
    """
    A model entry chooses a model by its key and which parameters to use.

    :see: :py:class:`pigreads.Models` and :py:class:`ModelDefinition`
    """

    key: str
    "The key of the model."
    parameters: MutableMapping[str, float] = Field(default_factory=dict)
    "The parameters of the model."
    model_config = ConfigDict(extra="forbid")
    "Configuration for :py:class:`pydantic.BaseModel`."

    @model_validator(mode="before")
    @classmethod
    def normalise(cls, values: Any) -> Any:
        """
        Normalise the model entry.

        If the entry is a string, it is converted to a dictionary with an empty
        parameter set. If the entry is a dictionary with a single key, the key
        is used as the model key.

        :param values: Any data to try to interpret.
        :return: Normalised data.
        """

        assert isinstance(values, dict)

        key = values.get("key")
        if "key" in values:
            del values["key"]

        parameters = values.get("parameters", {})
        if "parameters" in values:
            del values["parameters"]

        if key is None:
            assert len(values) == 1

            key, parameters = next(iter(values.items()))
            return {"key": key, "parameters": parameters or {}}

        parameters = {**parameters, **values}

        return {"key": key, "parameters": parameters or {}}


class ModelDefinition(BaseModel):
    """
    Definition of a so-called model, i.e., the reaction-term.

    Note: Additional fields are allowed stored in :py:attr:`meta`.

    :see: :py:class:`pigreads.Models` and :py:class:`ModelEntry`
    """

    name: str
    "Human readable name of the model, usually the authors and the year."
    description: str
    "A simple description of the model, what it was designed for, and its unique features."
    dois: list[str] = Field(default_factory=list)
    "List of digital object identifiers in URL form, i.e., starting with ``https://doi.org/10``."
    variables: dict[str, float]
    "Model variable names and their resting/initial values."
    diffusivity: dict[str, float]
    "Dictionary that maps the names of variables to diffuse to their diffusivity."
    parameters: dict[str, float]
    "Model parameter names and their default values."
    code: str
    "Source code of the forward Euler step in OpenCL."
    code_float: str | None = None
    "Source code for single precision floating point numbers, if none, use the generic source code."
    code_double: str | None = None
    "Source code for double precision floating point numbers, if none, use the generic source code."
    model_config = ConfigDict(extra="forbid")
    "Configuration for :py:class:`pydantic.BaseModel`."
    meta: dict[str, Any] = Field(default_factory=dict)
    "Additional fields as a dictionary."

    @model_validator(mode="before")
    @classmethod
    def normalise(cls, values: Any) -> dict[str, Any]:
        """
        Store additional fields as metadata.

        :see: :py:attr:`meta`
        """
        known_fields = set(cls.model_fields.keys())
        known = {k: v for k, v in values.items() if k in known_fields}
        unknown = {k: v for k, v in values.items() if k not in known_fields}
        known["meta"] = {**known.get("meta", {}), **unknown}
        return known

    @property
    def all_parameters(self) -> dict[str, float]:
        "Dictionary of parameters including the diffusivities."
        return {
            **{
                "diffusivity_" + varname: value
                for varname, value in self.diffusivity.items()
            },
            **self.parameters,
        }

    def __call__(self, **parameters: float) -> dict[str, float]:
        """
        Merge diffusivities, default parameters, and given parameters.

        :param parameters: Additional parameters.
        :return: Merged parameter dictionary.
        """
        return {**self.all_parameters, **parameters}

    def get_code(self, double_precision: bool = False) -> str:
        """
        Get the correct source code for a given floating point precision.

        :param double_precision: If ``True``, use double precision for calculations.
        :return: Source code of the forward Euler step in OpenCL.
        """
        if double_precision:
            if self.code_double is None:
                return self.code
            return self.code_double
        if self.code_float is None:
            return self.code
        return self.code_float



#!/bin/env python3
"""
Export Myokit MMT files to Pigreads code
----------------------------------------
"""

from __future__ import annotations

import sys
import warnings
from copy import deepcopy
from textwrap import dedent, indent
from typing import Any

import myokit as mk  # type: ignore[import-not-found]
import yaml
from myokit.formats.cellml import (  # type: ignore[import-not-found]
    CellMLImporter,
)
from myokit.formats.opencl import (  # type: ignore[import-not-found]
    OpenCLExpressionWriter,
)

INDENT: str = "    "


def str_presenter(dumper, data):
    """
    A YAML string presenter that uses block style for multiline strings.
    """
    data = data.strip()
    if len(data.splitlines()) > 1:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


yaml.representer.SafeRepresenter.add_representer(str, str_presenter)


class PigreadsMyokitWriter:
    """
    This class implements the conversion of a Myokit model to Pigreads
    compatible code.

    :param model: The Myokit model to convert.
    :param meta: A dictionary with metadata to include in the generated code.

    :var model: The Myokit model to convert.
    :vartype model: myokit.Model
    :var exwr1: The expression writer for single floating point precision
    :vartype exwr1: myokit.formats.opencl.OpenCLExpressionWriter
    :var exwr2: The expression writer for double floating point precision
    :vartype exwr2: myokit.formats.opencl.OpenCLExpressionWriter
    :var states: The list of states in the model.
    :vartype states: list[mk.State]
    :var meta: A dictionary with metadata to include in the generated code.
    :vartype meta: dict[str, Any]
    :var double_precision: If ``True``, use double precision for calculations.
    :vartype double_precision: bool
    """

    def __init__(self, model: mk.Model, meta: dict[str, Any]):
        self.model = model
        self.exwr1 = OpenCLExpressionWriter(
            precision=mk.SINGLE_PRECISION, native_math=True
        )
        self.exwr2 = OpenCLExpressionWriter(
            precision=mk.DOUBLE_PRECISION, native_math=False
        )
        self.exwr1.set_lhs_function(self.lhs_format)
        self.exwr2.set_lhs_function(self.lhs_format)
        self.states = list(model.states())
        self.meta = meta
        self.generate_variable_abbreviations()

    @property
    def diffusivities(self) -> dict[str, float]:
        """
        The diffusivities of the model as defined in the metadata.
        """
        d = self.meta.get("diffusivity", {})
        assert isinstance(d, dict)
        return d

    def get_ivar(self, varname: str) -> int:
        """
        Get the index of a state variable by its name.

        :param varname: The name of the state variable.
        :return: The index of the state variable.
        """
        return next(
            i
            for i, state in enumerate(self.states)
            if str(self.lhs_format(state)) == varname
        )

    @staticmethod
    def nodots(s: Any) -> str:
        """
        Convert an object to string and replace dots by underscores.

        :param s: any object to represent as a string.
        """
        return str(s).replace(".", "_")

    def lhs_format(self, x: mk.LhsExpression):
        """
        Format a left-hand side expression.

        :param x: The left-hand side expression to format.
        :return: The formatted left-hand side expression.
        """
        assert not isinstance(x, mk.Derivative), "Can not handle derivatives here."
        if isinstance(x, mk.Name):
            return self.lhs_format(x.var())
        s = self.nodots(x)
        return self.variable_abbreviations.get(s, s)

    @staticmethod
    def rush_larsen(
        v: mk.Variable,
        tau: mk.Variable,
        inf: mk.Variable,
        dt: str = "dt",
    ) -> mk.Expression:
        """
        The Rush-Larsen update for a state variable v.

        :param v: The state variable to update.
        :param tau: The time constant.
        :param inf: The steady state value.
        :param dt: Name of the time step.
        :return: The Rush-Larsen update for the state variable.
        """
        return mk.Plus(
            inf,
            mk.Multiply(
                mk.Minus(v, inf),
                mk.Exp(mk.PrefixMinus(mk.Divide(mk.Name(dt), tau))),
            ),
        )

    @staticmethod
    def safe_divide(a: mk.Expression, b: mk.Expression) -> mk.Expression:
        """
        Division that avoids division by a value close to zero.

        :param a: Enumerator.
        :param b: Denominator.
        :return: Expression of the safe division.
        """
        eps = mk.Name("VERY_SMALL_NUMBER")
        return mk.Divide(
            a,
            mk.If(
                i=mk.Less(mk.Abs(b), eps),
                t=mk.If(
                    i=mk.Less(b, mk.Number(0)),
                    t=mk.PrefixMinus(eps),
                    e=eps,
                ),
                e=b,
            ),
        )

    @classmethod
    def offset_in_division(cls, ex: mk.Expression) -> mk.Expression:
        """
        Avoid division by zero by adding a small offset in specific cases.

        :param ex: An expression to find and replace quotients in.
        :return: The updated expression.
        """

        subst: dict[Any, Any] = {}
        for quotient in ex.walk([mk.Divide]):
            numerator, denominator = quotient
            if not denominator.is_constant():
                subst[quotient] = cls.safe_divide(
                    cls.offset_in_division(numerator),
                    cls.offset_in_division(denominator),
                )
        return ex.clone(subst)

    def state_equation(self, q: mk.Equation, exwr: OpenCLExpressionWriter):
        """
        Format a state equation.

        :param q: The state equation to format.
        :return: The formatted state equation.
        """
        w = q.lhs.walk()
        next(w)
        v = next(w)
        vin = str(self.lhs_format(v))

        # if possible, use Rush-Larsen step
        if isinstance(q.rhs, mk.Divide):
            difference, tau = q.rhs
            if isinstance(difference, mk.Minus):
                left, right = difference
                if left == v:
                    return f"*_new_{vin} = -({exwr.ex(self.rush_larsen(v=v, tau=tau, inf=right))});"
                if right == v:
                    return f"*_new_{vin} = {exwr.ex(self.rush_larsen(v=v, tau=tau, inf=left))};"

        # else use forward Euler
        rhs: mk.Expression = q.rhs.clone()
        rhs = self.offset_in_division(rhs)
        update: str = str(exwr.ex(rhs))
        if vin in self.diffusivities:
            update += f" + _diffuse_{vin}"
        return f"*_new_{vin} = {vin} + dt*({update});"

    def generate_variable_abbreviations(self) -> None:
        """
        Create a dictionary abbreviating long variable names if this
        is possible unambiguously. Short variable names are the last part of the
        long variable name after a dot.
        """
        variables = dict(
            zip(self.model.states(), self.model.initial_values(as_floats=True))
        )
        variables_long = {self.nodots(v): f for v, f in variables.items()}
        variables_short = {}
        for variable, value in variables.items():
            short_varname = str(variable).rsplit(".", maxsplit=1)[-1]
            if short_varname not in variables_short:
                variables_short[short_varname] = value
            else:
                variables_short = variables_long
                break
        self.variables = variables_short
        self.variable_abbreviations = dict(
            zip(variables_long.keys(), variables_short.keys())
        )

    def to_dict(self) -> dict[str, Any]:
        """
        Convert the model to a dictionary. This is the main entry point for the
        conversion.

        :return: The model as a dictionary.
        """
        parameters = {
            self.lhs_format(q.lhs): float(q.rhs)
            for block in self.model.solvable_order().values()
            for q in block
            if q.rhs.is_constant()
        }

        d: dict[str, Any] = {
            "name": rf"{self.model.name()} (exported from Myokit)",
            "description": str(self.model.meta.get("desc", "")),
            "dois": [],
            "variables": self.variables,
            "parameters": parameters,
            **self.meta,
        }

        keys = ["code", "code_double"]
        for key, exwr in zip(keys, [self.exwr1, self.exwr2]):
            step = []
            for blockname, block in self.model.solvable_order().items():
                block_ = [eq for eq in block if not eq.rhs.is_constant()]
                if len(block_) < 1:
                    continue
                step.append("")
                step.append(f"// {blockname}")
                for q in block_:
                    if q.lhs.is_derivative():
                        step.append(self.state_equation(q, exwr))
                    else:
                        step.append(
                            "const Real "
                            + exwr.eq(
                                mk.Equation(q.lhs, self.offset_in_division(q.rhs))
                            )
                            + ";"
                        )
            d[key] = "\n".join(step).strip()

        if d[keys[0]] == d[keys[1]]:
            del d[keys[1]]

        return d

    def __str__(self) -> str:
        """
        Convert the model to a string. This is the another main entry point for the
        conversion.

        :return: The model as a string.
        """
        return yaml.safe_dump(
            self.to_dict(), sort_keys=False, indent=2, allow_unicode=True
        )


def main() -> None:
    from pathlib import Path

    cellml = CellMLImporter()

    path_models = Path(sys.argv[1])
    path_docs = Path(sys.argv[2])
    path_module = Path(sys.argv[3])

    paths_by_stem: dict[str, dict[str, str]] = {}
    for file in sorted(path_models.iterdir()):
        suffix = file.suffix[1:].lower()

        if suffix not in ["yaml", "mmt", "cellml"]:
            continue

        if file.stem not in paths_by_stem:
            paths_by_stem[file.stem] = {}

        paths_by_stem[file.stem][suffix] = str(file)

    models = {}
    for key, paths in paths_by_stem.items():
        model = {}
        meta = {}

        assert "yaml" in paths, "Must have yaml file."
        with Path(paths["yaml"]).open() as f:
            meta = yaml.safe_load(f.read())

        if "cellml" in paths:
            assert "mmt" not in paths, "Can only have mmt or cellml, not both."
            try:
                model = PigreadsMyokitWriter(
                    cellml.model(paths["cellml"]), meta
                ).to_dict()
            except Exception as e:
                warnings.warn(f"{paths['cellml']}: {e}", stacklevel=2)
                continue

        elif "mmt" in paths:
            try:
                model = PigreadsMyokitWriter(
                    mk.load_model(paths["mmt"]), meta
                ).to_dict()
            except Exception as e:
                warnings.warn(f"{paths['mmt']}: {e}", stacklevel=2)
                continue

        model.update(meta)
        models[key] = model

    path_docs.mkdir(exist_ok=True)
    for file in path_docs.iterdir():
        if file.is_file():
            file.unlink()

    warning: str = dedent("""
        AUTOMATICALLY GENERATED FILE!
        Edit ``src/models/compile.py`` or the model definitions in
        ``src/models/`` instead, run ``make`` in the main directory to compile.
    """).strip()

    with (path_docs / "index.rst").open("w") as f:

        def p(*args: Any) -> None:
            return print(*args, file=f)

        for line in warning.splitlines():
            p(f".. {line}")
        p()

        p(
            dedent(r'''
            Models
            ======

            A so-called model defines the reaction term of the reaction
            diffusion equation. While Pigreads comes with a variety of
            pre-defined models, it is also easily possible to define a
            model.

            Defining a model
            ----------------


            A model can be defined by adding it to the dictionary of available
            models::

                import pigreads as pig
                from pigreads.schema.model import ModelDefinition

                pig.Models.available["fitzhugh1961impulses"] = ModelDefinition(
                    name="FitzHugh 1961 & Nagumo 1962",
                    description="A 2D simplification of the Hodgkin-Huxley model.",
                    dois=[
                        "https://doi.org/10.1016/S0006-3495(61)86902-6",
                        "https://doi.org/10.1109/JRPROC.1962.288235",
                    ],
                    variables={"u": 1.2, "v": -0.625},
                    diffusivity={"u": 1.0},
                    parameters={"a": 0.7, "b": 0.8, "c": 3.0, "z": 0.0},
                    code="""
                        *_new_u = u + dt * (v + u - u*u*u/3 + z + _diffuse_u);
                        *_new_v = v + dt * (-(u - a + b*v)/c);
                    """,
                )


            The definition must adhere to the schema in :py:class:`pigreads.schema.model.ModelDefinition`.

            Pre-defined models
            ------------------

            .. toctree::
                :maxdepth: 1
                :hidden:
            ''').strip()
        )
        p()
        for key in models:
            p(f"    {key}")
        p()
        p(
            dedent(r"""
            .. csv-table::
                :header: Name, Key, Variables, Parameters
            """).strip()
        )
        p()
        for key, model in models.items():
            p(
                f'    ":doc:`{model["name"]} <{key}>`", "``{key}``", {len(model["variables"])}, {len(model["parameters"])}'
            )

    for key, model in deepcopy(models).items():
        with (path_docs / f"{key}.rst").open("w") as f:

            def p(*args: Any) -> None:
                return print(*args, file=f)

            def code(s: str, summary: str) -> None:
                p(".. raw:: html")
                p()
                p("    <details>")
                p(f"    <summary>{summary}</summary>")
                p()
                p(".. code-block:: c")
                p()
                p(indent(s.strip(), INDENT))
                p()
                p(".. raw:: html")
                p()
                p("    </details>")

            for line in warning.splitlines():
                p(f".. {line}")
            p()

            name: str = model.pop("name")
            p(name)
            p("=" * len(name))
            p()
            p(f"**Key:** ``{key}``")
            p()
            p(model.pop("description").strip())
            p()
            dois: list[str] = model.pop("dois")
            if len(dois) > 0:
                p("References")
                p("----------")
                p()
                for i, doi in enumerate(dois):
                    p(f"{i}. {doi}")
                p()
            p("Variables")
            p("---------")
            p()
            for i, (name, value) in enumerate(model.pop("variables").items()):
                p(f"{i}. ``{name} = {value}``")
            p()
            p("Parameters")
            p("----------")
            p()
            for name, value in model.pop("diffusivity").items():
                p(f"- ``diffusivity_{name} = {value}``")
            for name, value in model.pop("parameters").items():
                p(f"- ``{name} = {value}``")
            p()
            p("Source code")
            p("-----------")
            p()
            code(model.pop("code"), "General code")
            if model.get("code_float", None) is not None:
                p()
                code(model.pop("code_float"), "Code for single precision")
            if model.get("code_double", None) is not None:
                p()
                code(model.pop("code_double"), "Code for double precision")
            if len(model) > 0:
                p()
                p("Additional metadata")
                p("-------------------")
                p()
                p(".. code-block:: yaml")
                p()
                p(
                    indent(
                        yaml.safe_dump(
                            model, sort_keys=False, indent=2, allow_unicode=True
                        ),
                        INDENT,
                    )
                )

    with path_module.open("w") as f:

        def p(*args: Any) -> None:
            return print(*args, file=f)

        for line in warning.splitlines():
            p(f"# {line}")
        p()
        p(
            dedent(f'''
            """
            Pre-defined models
            ------------------
            """

            from typing import Any

            available: dict[str, dict[str, Any]] = {models!r}
        ''').strip()
        )


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
import re
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
import pytest
import yaml
from click.testing import CliRunner

from pigreads import __version__
from pigreads.cli import main as cli
from pigreads.progress import PROGRESS_ITERS, BaseProgressIterator
from pigreads.schema import Simulation

yaml_data = """
pigreads: 1
Nfr: 3
Nt: 2
Nz: 1
Ny: 1
Nx: 1
dt: 0.1
dz: 0.1
dy: 0.1
dx: 0.1
diffusivity: 0.1
models: aliev1996simple
init:
    u: 0.3
"""
sim = Simulation(**yaml.safe_load(yaml_data))


def test_usage():
    runner = CliRunner()
    result = runner.invoke(cli)
    print(result.output)
    assert result.exit_code in [0, 2]
    assert result.output.startswith("Usage")


def test_version():
    runner = CliRunner()
    result = runner.invoke(cli, ["version"])
    print(result.output)
    assert result.exit_code == 0
    assert result.stdout.strip() == f"Pigreads {__version__}"


def test_yaml2json():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.yaml"
        path_output = path / "output.json"

        with path_input.open("w") as f:
            f.write(yaml_data)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["convert", "-f", "json", str(path_input), str(path_output)],
        )
        assert result.exit_code == 0

        with path_output.open() as f:
            assert Simulation(**json.loads(f.read())) == sim


def test_json2yaml():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.json"
        path_output = path / "output.yaml"

        with path_input.open("w") as f:
            f.write(sim.model_dump_json())

        runner = CliRunner()
        result = runner.invoke(
            cli,
            ["convert", str(path_input), str(path_output)],
        )
        assert result.exit_code == 0

        with path_output.open() as f:
            assert Simulation(**yaml.safe_load(f.read())) == sim


def test_invalid_conversion():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.json"
        path_output = path / "output.yaml"

        with path_input.open("w") as f:
            f.write(sim.model_dump_json())

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "convert",
                "-f",
                "invalid",
                str(path_input),
                str(path_output),
            ],
        )
        assert result.exit_code == 2
        assert "Invalid value" in result.output


def test_run():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.yaml"
        path_output = path / "output.npy"

        with path_input.open("w") as f:
            f.write(yaml_data)

        runner = CliRunner()
        result = runner.invoke(cli, ["run", str(path_input), str(path_output)])
        assert result.exit_code == 0

        states = np.load(path_output)
        assert np.allclose(states[-1].squeeze(), [0.4401616, 0.00194737])


def test_live():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.yaml"
        path_output = path / "output.npy"
        path_live = path / "live.png"

        with path_input.open("w") as f:
            f.write(yaml_data)

        runner = CliRunner()
        result = runner.invoke(
            cli,
            [
                "run",
                str(path_input),
                str(path_output),
                "--live",
                "u",
                "--live-path",
                str(path_live),
                "--live-kwargs={style: dark_background}",
            ],
        )
        assert result.exit_code == 0

        states = np.load(path_output)
        assert np.allclose(states[-1].squeeze(), [0.4401616, 0.00194737])


def test_progress():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.yaml"
        path_output = path / "output.npy"

        with path_input.open("w") as f:
            f.write(yaml_data)

        for progress in PROGRESS_ITERS:
            runner = CliRunner()
            result = runner.invoke(
                cli, ["run", "-p", progress, str(path_input), str(path_output)]
            )
            assert result.exit_code == 0


def test_progress_base():
    for _ in BaseProgressIterator(range(3)):
        pass


def test_movie():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.yaml"
        path_output = path / "output.npy"
        path_movie = path / "movie.mp4"

        with path_input.open("w") as f:
            f.write(yaml_data)

        runner = CliRunner()
        result = runner.invoke(cli, ["run", str(path_input), str(path_output)])
        assert result.exit_code == 0

        result = runner.invoke(
            cli, ["movie", str(path_input), str(path_output), str(path_movie)]
        )
        assert result.exit_code == 0


def test_unstable():
    with TemporaryDirectory() as tempdir:
        path = Path(tempdir)
        path_input = path / "input.yaml"
        path_output = path / "output.npy"

        with path_input.open("w") as f:
            f.write(re.sub(r"dt:.*", "dt: 1e10", yaml_data))

        runner = CliRunner()
        with pytest.warns(RuntimeWarning, match="New non-finite values found"):
            result = runner.invoke(cli, ["run", str(path_input), str(path_output)])
        assert result.exit_code == 0
